
//
// flic (.flc) file viewer
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <Ph.h>
#include <Pt.h>
#include "flic.h"

//
// memcpy/set routines
//

memcpy8( void * from, void * to, int count );
#pragma aux memcpy8 =\
	"rep movsb" \
parm nomemory [esi] [edi] [ecx] \
modify exact [esi edi ecx];
	
memcpy16( void * from, void * to, int count );
#pragma aux memcpy16 =  \
	"rep movsw" \
parm nomemory [esi] [edi] [ecx] \
modify exact [esi edi ecx];

memset8( void * to, int value, int count );
#pragma aux memset8 =\
	"rep stosb" \
parm nomemory [edi] [al] [ecx] \
modify exact [edi al ecx];

memset16( void * to, int value, int count );
#pragma aux memset16 =\
	"rep stosw" \
parm nomemory [edi] [ax] [ecx] \
modify exact [edi ax ecx];

//
// globals
//

Flic * gflic;
int debug;
int shmem;
int one;

//
// functions
//

// load flic file
Flic * FlicLoad( int fd )
{
	U4 size;
	U2 type;
	Header * header;
	char * buffer;
	Flic * flic;
	Frame * * frames;
	Frame * frame;
	int i;

	// read file
	read( fd, & size, 4 );
	read( fd, & type, 2 );
	if( type != TYPE_FLC && type != TYPE_FLI )
		return( NULL );
	// flic
	flic = calloc( 1, sizeof( Flic ) );
	if( ! flic )
		return( NULL );
	// buffer
	if( shmem )
		buffer = PgShmemCreate( size, NULL );
	else
		buffer = malloc( size );
	if( ! buffer )
	{
		FlicDestroy( flic );
		return( NULL );
	}
	flic->buffer = buffer;
	// read file
	if( read( fd, buffer + 6, size - 6 ) != ( size - 6 ) )
	{
		FlicDestroy( flic );
		return( NULL );
	}
	// header
	header = ( Header * ) buffer;
	header->size = size;
	header->type = type;
	if( debug )
		HeaderPrint( header );
	flic->header = header;
	// skip prefix
	// speed
	if( type == TYPE_FLI )
		header->speed = max( header->speed * 14, 1 );
	else
		header->speed = max( header->speed, 1 );
	// frames
	if( one )
		header->frames = 1;
	if( header->flags & FLAG_RING )
		header->frames++; // add one for ring frame
	frames = calloc( header->frames, sizeof( Frame * ) );
	if( ! frames )
	{
		FlicDestroy( flic );
		return( NULL );
	}
	flic->frames = frames;
	if( header->type == TYPE_FLC )
		frame = ( Frame * ) ( buffer + header->oframe1 );
	else
		frame = ( Frame * ) ( buffer + sizeof( Header ) );
	for( i = 0; i < header->frames; i++ )
	{
		if( frame->type == TYPE_FRAME )
		{
			frames[i] = frame;
		}
		else
		{
			if( debug )
				printf( "flc: incorrect frame type\n" );
		}
		frame = ( Frame * ) ( ( char * ) frame + frame->size );
	}
	return( flic );
}

void FlicDestroy( Flic * flic )
{
	if( ! flic )
		return;
	if( flic->buffer )
		if( shmem )
			PgShmemDestroy( flic->buffer );
		else
			free( flic->buffer );
	if( flic->frames )
		free( flic->frames );
	if( flic->image )
	{
		if( flic->image->image )
			if( shmem )
				PgShmemDestroy( flic->image->image );
			else
				free( flic->image->image );
		if( flic->image->palette )
			free( flic->image->palette );
		free( flic->image );
	}
	free( flic );
}

// initialize flic for display
void FlicInit( Flic * flic )
{
	PhImage_t * image;
	char * data;
	PgColor_t * palette;
	int n;

	image = calloc( 1, sizeof( PhImage_t ) );
	if( ! image )
		return;
	n = flic->header->height * flic->header->width;
	if( shmem )
		data = PgShmemCreate( n, NULL );
	else
		data = calloc( n, sizeof( char ) );
	if( ! data )
		return;
	palette = calloc( 256, sizeof( PgColor_t ) );
	if( ! palette )
		return;
	image->type = Pg_IMAGE_PALETTE_BYTE;
	image->bpl = flic->header->width;
	image->size.h = flic->header->height;
	image->size.w = flic->header->width;
	image->colors = 256;
	image->image = data;
	image->palette = palette;
	flic->image = image;
}

// display the frame
void DecodeFrame( Flic * flic, Frame * frame )
{
	int i;
	Chunk * chunk;

	chunk = ( Chunk * ) ( ( char * ) frame + sizeof( Frame ) );
	for( i = 0; i < frame->chunks; i++ )
	{
		switch( chunk->type )
		{
			case TYPE_BYTE_RUN:
				DecodeByteRun( flic, chunk );
				break;
			case TYPE_COLOR_256:
				DecodeColor256( flic, chunk );
				break;
			case TYPE_COLOR_64:
				DecodeColor64( flic, chunk );
				break;
			case TYPE_DELTA_FLC:
				DecodeDeltaFlc( flic, chunk );
				break;
			case TYPE_DELTA_FLI:
				DecodeDeltaFli( flic, chunk );
				break;
			case TYPE_FLI_COPY:
				DecodeFliCopy( flic, chunk );
				break;
			case TYPE_BLACK:
				DecodeBlack( flic, chunk );
				break;
			case TYPE_PSTAMP:
			default:
				if( debug )
					printf( "flc: unsupported chunk type %d\n", chunk->type );
				break;
		}
		chunk = ( Chunk * ) ( ( char * ) chunk + chunk->size );
	}
}

void DecodeColor256( Flic * flic, Chunk * chunk )
{
	int packets;
	int i, j, n;
	int skip;
	int count;
	int index;
	int r, g, b;
	char * p;

	p = ( char * ) chunk + sizeof( Chunk );
	packets = * ( U2 * ) p;
	p += 2;
	//printf( "packets %d\n", packets );
	index = 0;
	for( i = 0; i < packets; i++ )
	{
		skip = * p++;
		//printf( "skip %u\n", skip );
		count = * p++;
		if( count == 0 )
			n = 256;
		else
			n = count;
		//printf( "count %u\n", n );
		index += skip;
		for( j = 0; j < n; j++ )
		{
			r = * p++;
			g = * p++;
			b = * p++;
			flic->image->palette[index++] = PgRGB( r, g, b );
		}
	}
}

void DecodeColor64( Flic * flic, Chunk * chunk )
{
	int packets;
	int i, j, n;
	int skip;
	int count;
	int index;
	int r, g, b;
	char * p;

	p = ( char * ) chunk + sizeof( Chunk );
	packets = * ( U2 * ) p;
	p += 2;
	//printf( "packets %d\n", packets );
	index = 0;
	for( i = 0; i < packets; i++ )
	{
		skip = * p++;
		//printf( "skip %u\n", skip );
		count = * p++;
		if( count == 0 )
			n = 256;
		else
			n = count;
		//printf( "count %u\n", n );
		index += skip;
		for( j = 0; j < n; j++ )
		{
			r = ( * p++ ) << 2;
			g = ( * p++ ) << 2;
			b = ( * p++ ) << 2;
			flic->image->palette[index++] = PgRGB( r, g, b );
		}
	}
}

void DecodeByteRun( Flic * flic, Chunk * chunk )
{
	int x, y;
	char * p;
	char * q;
	S1 b;
	int n;
	int i;

	p = ( char * ) chunk + sizeof( Chunk );
	x = 0;
	q = flic->image->image;
	while( x < flic->header->height )
	{
		p++; // skip first byte
		y = 0;
		while( y < flic->header->width )
		{
			// type byte
			b = * p++;
			if( b >= 0 )
			{
				// replicate 
 				n = b;
				memset8( q, * p, n );
				q += n;
				p++;
			}
			else
			{
				// read
				n = -b;
				memcpy8( p, q, n );
				q += n;
				p += n;
			}
			y += n;
		}
		x++;
	}
}

void DecodeFliCopy( Flic * flic, Chunk * chunk )
{
	char * p;
	char * q;

	p = ( char * ) chunk + sizeof( Chunk );
	q = flic->image->image;
	memcpy8( p, q, flic->header->height * flic->header->width );
}

void DecodeBlack( Flic * flic, Chunk * chunk )
{
	char * q;

	q = flic->image->image;
	memset8( q, 0, flic->header->height * flic->header->width );
}

void DecodeDeltaFlc( Flic * flic, Chunk * chunk )
{
	int lines;
	char * p;
	char * q;
	int x, y;
	int i, j, k, line;
	U1 skip;
	U1 n;
	S1 type;
	S2 packet;
	U2 opcode;

	y = 0;
	p = ( char * ) chunk + sizeof( Chunk );
	lines = * ( U2 * ) p;
	//printf( "lines %d\n", lines );
	p += 2;
	for( i = 0; i < lines; i++ )
	{
		x = 0;
		while( 1 )
		{	
			packet = * ( U2 * ) p;
			p += 2;
			opcode = packet & 0xC000;
			if( opcode == 0x0000 )
			{
				// get packets
				for( j = 0; j < packet; j++ )
				{
					skip = * ( U1 * ) p;
					p++;
					x += skip;
					//printf( "column skip %u\n", skip );
					type = * ( S1 * ) p;
					p++;
					q = flic->image->image + ( flic->header->width * y ) + x;
					if( type < 0 )
					{
						// replicate 
 						n = -type;
						//printf( "replicate %u\n", n );
						memset16( q, * ( U2 * ) p, n );
						p += 2;
					}
					else
					{
						// read
						n = type;
						//printf( "copy %u\n", type );
						memcpy16( p, q, n );
						p += ( n * 2 );
					}
					//printf( "x=%d y=%d s=%d n=%d\n", x, y, skip, n * 2 );
					x += ( n * 2 );
				}
				y++;
				break;
			}
			else if( opcode == 0x8000 )
			{
				printf( "last byte\n" );
				flic->image->image[ y * flic->header->width + flic->header->width - 1 ] = ( packet & 0x00FF );
				continue;
			}
			else if( opcode == 0xC000 )
					{
				//printf( "line skip %d\n", - packet );
				y -= packet;
				continue;
			}
			else
			{
				if( debug )
					printf( "flc: erroneous opcode\n" ); 
				break;
			}
		}
	}
}

void DecodeDeltaFli( Flic * flic, Chunk * chunk )
{
	char * p;
	char * q;
	int x, y;
	int i, j, k;
	U2 lines;
	U1 skip;
	U1 n;
	S1 type;
	U1 packets;

	p = ( char * ) chunk + sizeof( Chunk );
	y = * ( U2 * ) p;
	p += 2;
	//printf( "line skip %u\n", y );
	lines = * ( U2 * ) p;
	p += 2;
	//printf( "lines %u\n", lines );
	for( i = 0; i < lines; i++ )
	{
		x = 0;
		packets = * ( U1 * ) p;
		p++;
		//printf( "packets %u\n", packets );
		for( j = 0; j < packets; j++ )
		{
			skip = * ( U1 * ) p;
			p++;
			x += skip;
			//printf( "column skip %u\n", skip );
			type = * ( S1 * ) p;
			p++;
			q = flic->image->image + ( flic->header->width * y ) + x;
			if( type < 0 )
			{
				// replicate 
				n = -type;
				//printf( "replicate %u\n", n );
				memset8( q, * p, n );
				p++;
			}
			else
			{
				// read
				n = type;
				//printf( "copy %u\n", type );
				memcpy8( p, q, n );
				p += n;
			}
			x += n;
		}
		y++;
	}
}

void HeaderPrint( Header * header )
{
	printf( "size = %lu\n", header->size );
	printf( "type = 0x%hX\n", header->type );
	printf( "frames = %hu\n", header->frames );
	printf( "width = %hu\n", header->width );
	printf( "height = %hu\n", header->height );
	printf( "depth = %hu\n", header->depth );
	printf( "flags = 0x%hX\n", header->flags );
	printf( "speed = %lu\n", header->speed );
	printf( "created = 0x%hX\n", header->created );
	printf( "creator = 0x%hX\n", header->creator );
	printf( "updated = 0x%hX\n", header->updated );
	printf( "updator = 0x%hX\n", header->updator );
	printf( "aspect_dx = %hu\n", header->aspect_dx );
	printf( "aspect_dy = %hu\n", header->aspect_dy );
	printf( "oframe1 = %lu\n", header->oframe1 );
	printf( "oframe2 = %lu\n", header->oframe2 );
}

int display( void * data )
{
	PtWidget_t * widget = ( PtWidget_t * ) data;
	PtArg_t	args[1];

	//printf( "frame %d\n", gflic->index );
	DecodeFrame( gflic, gflic->frames[gflic->index] );
	PtSetArg( & args[0], Pt_ARG_LABEL_DATA,	gflic->image, sizeof( PhImage_t ) );
	PtSetResources( widget, 1, args );
	PtFlush();
	gflic->index++;
	if( gflic->index == gflic->header->frames )
		if( gflic->header->flags & FLAG_INITIAL )
			gflic->index = 1;
		else
			gflic->index = 0;
	return( Pt_CONTINUE );
}

int tick( PtWidget_t * widget, void * data, PtCallbackInfo_t * cbinfo )
{
	if( ! one )
		PtTimerArm( widget, gflic->header->speed );
	display( widget );
	return( Pt_CONTINUE );
}

void main( int argc, char * * argv )
{
	PtAppContext_t app;
	PtWidget_t * label;
	PtArg_t	args[10];
	PhPoint_t pos;
	PhDim_t dim;
	PtWidget_t * window;
	Flic * flic;
	int fd;
	int n;
	int c;
	PtRawCallback_t callback = { Ph_EV_TIMER, tick, NULL };

	debug = 0;
	shmem = 1;
	one = 0;
	while( ( c = getopt( argc, argv, "ds1" ) ) != -1 )
	{
		switch( c )
		{
			case 'd':
				debug++;
				break;
			case 's':
				shmem = 0;
				break;
			case '1':
				one++;
				break;
		}
	}
	fd = open( argv[optind], O_RDONLY );
	if( fd < 0 )
		return;
	flic = FlicLoad( fd );
	if( ! flic )
		return;
	FlicInit( flic );
	gflic = flic;
	// create window
	if( ( window = PtAppInit( & app, & argc, argv, 0, NULL ) ) == NULL )
		return;
	n = 0;
	PtSetArg( & args[n], Pt_ARG_RESIZE_FLAGS, Pt_RESIZE_XY_ALWAYS, 0 ); n++;
	PtSetArg( & args[n], Pt_ARG_WINDOW_TITLE, argv[optind], 0 ); n++;
	PtSetResources( window, n, args );
	// create label
	pos.x = 0;
	pos.y = 0;
	dim.h = flic->header->height;
	dim.w = flic->header->width;
	n = 0;
	PtSetArg( & args[n], Pt_ARG_FLAGS, Pt_FALSE, Pt_HIGHLIGHTED ); n++;
	PtSetArg( & args[n], Pt_ARG_POS, & pos, 0 ); n++;
	PtSetArg( & args[n], Pt_ARG_DIM, & dim, 0 ); n++;
	PtSetArg( & args[n], Pt_ARG_BORDER_WIDTH, 0, 0 ); n++;
	PtSetArg( & args[n], Pt_ARG_MARGIN_HEIGHT, 0, 0 ); n++;
	PtSetArg( & args[n], Pt_ARG_MARGIN_WIDTH, 0, 0 ); n++;
	PtSetArg( & args[n], Pt_ARG_LABEL_TYPE, Pt_IMAGE, 0 ); n++;
	PtSetArg( & args[n], Pt_CB_RAW, & callback, 0 ); n++;
	label = PtCreateWidget( PtLabel, window, n, args );
	PtRealizeWidget( window );
//	PtAppAddWorkProc( NULL, & display, label );
	PtTimerArm( label, flic->header->speed );
	PtMainLoop();
	FlicDestroy( gflic );
}

